// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Handler_hpp
#define anna_comm_Handler_hpp

#include <netinet/in.h>

#include <anna/core/mt/Runnable.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/util/Microsecond.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Communicator;
class Transport;
class ClientSocket;

/**
   Controlador de comunicaciones generico.
*/
class Handler : public Runnable {
public:
  using Runnable::initialize;
  using Runnable::run;
  using Runnable::setIsRunning;

  /**
     Mascara que define el funcionamiento de los manejadores de mensajes.
  */
  struct Support { enum _v { None = 0, CongestionControl = 1 }; };

  /**
     Tipo de controladores predefinidos.
  */
  struct Type {
    enum _v {
      ServerSocket,  /**< Controlador para ServerSocket */
      LocalConnection,  /**< Controlador para LocalConnection */
      RemoteConnection,  /**< Controlador para RemoteConnection */
      DatagramSocket,  /**< Controlador para DatagramSocket */
      BinderSocket,  /**< Controlador para comm::BinderSocket usado para compartir direcciones IPs. */
      Custom,  /**< Controlador definido por el usuario */
      ClientSocket /**< Controlador para un ClientSocket directo (sin uso intermedio de Server */
    };
  };

  /**
     Devuelve el tipo de controlador.
     \return el tipo de controlador.
  */
  Type::_v getType() const throw() { return a_type; }

  /**
     Devuelve el descriptor de fichero asociado a este controlador.
     \return el descriptor de fichero asociado a este controlador.
  */
  int getfd() const throw() { return a_fd; }

  /**
     Devuelve \em true si el descriptor de fichero asociado a este controlador soporta
     control de congestion o \em false en otro caso.
     \return \em true si el descriptor de fichero asociado a este controlador soporta
     control de congestion o \em false en otro caso.
  */
  bool supportCongestionControl() const throw() { return (a_support & Support::CongestionControl) != 0; }

  /**
     Devuelve \em true si este manejador soporta control de temporizacion o \em false en otro
     caso.
     \return \em true si este manejador soporta control de temporizacion o \em false en otro
     caso.
  */
  bool supportTimeout() const throw() { return a_timeout > 0; }

  /**
     Operador de comparacion.
     \param fd Descriptor de fichero con el que comparar.
     \return \em true si el fd recibido es igual al establecido con #setfd o \em false en caso contrario.
  */
  bool operator == (const int fd) const throw() { return a_fd == fd; }

  /**
    Metodo invocado por el comunicador cuando detectado actividad en el descriptor de
    fichero asociado a este controlador.
  */
  virtual void apply() throw(RuntimeException) = 0;

  /**
     Devuelve el ClientSocket asociado a este manejador de conexiones.
     \return El ClientSocket asociado a este manejador de conexiones. Puede ser NULL.
     \warning Uso interno. Se necesita para poder cooperar con el anna::comm::CongestionController.
  */
  virtual ClientSocket* getClientSocket() throw() { return NULL; }

  /**
     Devuelve una cadena con la informacion referente a esta instancia.
     \return una cadena con la informacion referente a esta instancia.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion referente a esta instancia.
     \return un documento XML con la informacion referente a esta instancia.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

  /**
     Amplia la informacion XML del nodo recibido como parametro.
     \param node Nodo XML en el que incorporar los atributos.
  */
  void asAttribute(xml::Node* node) const throw(RuntimeException);

protected:
  /**
     Instancia del comunicador puede ser NULL.
  */
  Communicator* a_communicator;

  /**
     Constructor.
     \param communicator Comunicador asociado a este controlador.
     \param type Tipo de Comunicador.
     \param support Una combinacion de los valores de Handler::Support.
  */
  Handler(Communicator* communicator, const Type::_v type, const int support = Support::CongestionControl) :
    a_communicator(communicator),
    a_type(type),
    a_support(support),
    a_fd(-1),
    a_timeout(0),
    a_maxTime(0),
    a_loop(0)
  {;}

  /**
     Constructor.
     \param type Tipo de Comunicador.
     \param support Una combinacion de los valores de Handler::Support.
  */
  Handler(const Type::_v type, const int support = Support::CongestionControl) :
    a_communicator(NULL),
    a_type(type),
    a_support(support),
    a_fd(-1),
    a_timeout(0),
    a_maxTime(0),
    a_loop(0)
  {;}

  /**
     Establecer el descriptor de fichero asociado a este controlador.
     \param fd Descriptor de fichero asociado a este controlador.
     \warning La implementacion del metodo initialize debe invocar a este metodo
     con descriptor de fichero valido.
  */
  void setfd(const int fd) throw() { setId(anna::functions::asText("Handler", a_fd = fd)); }

  /**
     Establece el numero de milisegundos maximo que puede estar este manejador sin
     recibir mensajes antes de ser cerrado por el nucleo.
     \param timeout Numero de milisegundos maximo sin recibir mensajes.
  */
  void setTimeout(const Millisecond &timeout) throw() {
    a_timeout = timeout;
    a_maxTime = functions::hardwareClock() + a_timeout;
  }

  /**
     En los manejadores que pueden recibir mas de una peticion en cada llamada a apply este
     metodo debe ser invocado para saber si debe dejar de procesar mensajes.
  */
  bool canContinue() const throw() { return hasRequestedStop() == false; }

  /**
     Metodo con el que podemos redefinir el comportamiento cuando recibe la notificacion de
     que una IP ha dejado de estar disponible.
     \param address Direccion IP que ha dejado de estar disponible.
     \warning Se invoca automaticamente desde el comunicador.
  */
  virtual void breakAddress(const in_addr_t& address) throw() {;}

  /**
     Metodo con el que podemos redefinir el comportamiento cuando recibe la notificacion de
     que una IP esta disponible.
     \param address Direccion IP que ha pasado a estar disponible.
     \warning Se invoca automaticamente desde el comunicador.
  */
  virtual void recoverAddress(const in_addr_t& address) throw() {;}

  /**
   * Mtodo que se invoca peridicamente para comprobar si tenemos pendiente el cierre de la conexin
   * con el canal asociado a este manejador, cuando el fd asociado al manejador de recibe actividad.
   *
   * \return \em true Si termina la conexin o \em false en otro caso.
   */
  virtual bool testClose() throw(RuntimeException) { return false;}

  /**
     Metodo con el que podemos redefinir el comportamiento cuando recibe la notificacion de
     que el componente asociado a este controlador ha dejado de estar operativo.
     \warning Se invoca automaticamente desde el comunicador al invocar al metodo
     \em detach correspondiente.
  */
  virtual void finalize() throw() {;}

  /**
   * Metodo con el que podemos redefinir el comportamiento cuando recibe la notificacion de
   * que el componente asociado a este controlador ha sido duplicado en un proceso hijo.
   * \warning Exclusivamente uso interno.
  */
  virtual void clone() throw(RuntimeException) {;}

private:
  const Type::_v a_type;
  const int a_support;
  int a_fd;
  Microsecond a_timeout;
  Microsecond a_maxTime;
  int a_loop;

  Handler(const Handler&);
  void do_action() throw(RuntimeException);
  void beat(const Microsecond& now) throw() { a_maxTime = now + a_timeout; }
  bool isTimeout(const Microsecond& now) { return a_maxTime > 0 && a_maxTime <= now; }

  friend class Communicator;
};

}
}

#endif

