// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_Second_hpp
#define anna_core_util_Second_hpp

#include <string>

#include <anna/config/defines.hpp>
#include <anna/core/RuntimeException.hpp>

namespace anna {

class Millisecond;
class Microsecond;

/**
 * Clase para modelar la unidad temporal segundos.
 */
class Second {
public:
#ifdef __anna64__
  typedef S64 type_t;
#else
  typedef int type_t;
#endif

  /**
     Tamao de la memoria reservada que debe tener la variable usada para guardar
     el resultado de convertir el 'time' en texto.

     @see #asDateTime
  */
  static const int DateTimeSizeString = 21;

  /**
   * Constructor
   */
  Second() : a_value(0) {;}

  /**
   * Constructor.
   * \param value Valor inicial de esta instancia.
   */
  explicit Second(const type_t value) : a_value(value) {;}

  /**
   * Constructor copia.
   * \param other Instancia de la que copiar.
   */
  Second(const Second& other) : a_value(other.a_value) {;}

  /**
   * Constructor copia.
   * \param other Instancia de la que copiar.
   */
  Second(const Millisecond& other);

  /**
   * Constructor copia.
   * \param other Instancia de la que copiar.
   */
  Second(const Microsecond& other);

  /**
   * Conversor a numrico.
   * \return El valor asociado a esta instancia.
   */
  operator type_t () const throw() { return a_value; }

  /**
   * \internal
   */
  type_t& refValue() throw() { return a_value; }

  Second& operator= (const type_t other) throw() { a_value = other; return *this; }

  Second& operator= (const Second& other) throw() { a_value = other.a_value; return *this; }

  Second& operator= (const Millisecond& other) throw();

  Second& operator= (const Microsecond& other) throw();

  bool operator== (const Second& other) const throw() { return a_value == other.a_value; }

  bool operator== (const Millisecond& other) const throw();

  bool operator== (const Microsecond& other) const throw();

  bool operator!= (const Second& other) const throw() { return a_value != other.a_value; }

  bool operator!= (const Millisecond& other) const throw();

  bool operator!= (const Microsecond& other) const throw();

  bool operator> (const Second& other) const throw() { return a_value > other.a_value; }

  bool operator> (const Millisecond& other) const throw();

  bool operator> (const Microsecond& other) const throw();

  bool operator< (const Second& other) const throw() { return a_value < other.a_value; }

  bool operator< (const Millisecond& other) const throw();

  bool operator< (const Microsecond& other) const throw();

  bool operator>= (const Second& other) const throw() { return a_value >= other.a_value; }

  bool operator>= (const Millisecond& other) const throw() { return (operator==(other) == true) ? true : operator>(other); }

  bool operator>= (const Microsecond& other) const throw() { return (operator==(other) == true) ? true : operator>(other); }

  bool operator<= (const Second& other) const throw() { return a_value <= other.a_value; }

  bool operator<= (const Millisecond& other) const throw() { return (operator==(other) == true) ? true : operator<(other); }

  bool operator<= (const Microsecond& other) const throw() { return (operator==(other) == true) ? true : operator<(other); }

  /**
   * Devuelve el valor asociado a esta instancia.
   * \return el valor asociado a esta instancia.
   */
  type_t getValue() const throw() { return a_value; }

  /**
     Devuelve una cadena con la hora en formato 'dd/mm/yyyy hh:mm:ss'.
     @param format Indicador de formato.
     @return Un literal con la hora en el formato 'dd/mm/yyyy hh:mm:ss'.

     \see man strftime
  */
  std::string asDateTime(const char* format = "%d/%0m/%Y %T") const throw();

  /**
      Devuelve una cadena con la hora en formato 'dd/mm/yyyy hh:mm:ss'.

      @param format Indicador de formato.
      @param result Puntero donde vamos a guardar el resultado de la conversin.
      Debe tener espacio reservado para contener #DateTimeSizeString caracteres.

      \see man strftime
   */
  const char* asDateTime(char* result, const char* format = "%d/%0m/%Y %T") const throw();

  /**
   * Devuelve una cadena con el valor de esta instancia y las unidades "sec".
   * \return una cadena con el valor de esta instancia y las unidades "sec".
   */
  std::string asString() const throw();

  /**
   * Devuelve la hora actual de sistema expresada en segundos transcurridos desde el 1 de Enero de 1970
   * \return la hora actual de sistema expresada en segundos transcurridos desde el 1 de Enero de 1970
   */
  static Second getTime() throw();

  /**
   * Devuelve la hora actual de sistema expresada en segundos transcurridos desde el 1 de Enero de 1970 aplicando
   * las correciones correspondientes a la hora local.
   * \return la hora actual de sistema expresada en segundos transcurridos desde el 1 de Enero de 1970
   */
  static Second getLocalTime() throw();

  /**
   * Obtiene los microsegundos del valor contenido en la cadena recibida como parmetro.
   * \param value Cadena que contiene los microsegundos habr sido obtenida con #asString.
   * \return los microsegundos del valor contenido en la cadena recibida como parmetro.
   */
  static Second fromString(const std::string& value) throw(RuntimeException);

private:
  type_t a_value;

  friend class Millisecond;
  friend class Microsecond;

  friend class Second operator + (const Second& left, const Second& right) throw();
  friend class Second operator - (const Second& left, const Second& right) throw();
  friend class Second operator / (const Second& left, const Second& right) throw();
  friend class Second operator / (const Second& left, const int right) throw();
  friend class Second operator / (const Second& left, const unsigned int right) throw();
};

inline Second operator + (const Second& left, const Second& right)
throw() {
  return Second(left.a_value + right.a_value);
}

inline Second operator - (const Second& left, const Second& right)
throw() {
  return Second(left.a_value - right.a_value);
}

inline Second operator / (const Second& left, const Second& right)
throw() {
  return Second(left.a_value / right.a_value);
}

inline Second operator / (const Second& left, const int right)
throw() {
  return Second(left.a_value / right);
}

inline Second operator / (const Second& left, const unsigned int right)
throw() {
  return Second(left.a_value / right);
}

}

#endif
