// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_statistics_Accumulator_hpp
#define anna_statistics_Accumulator_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/functions.hpp>

// Standard
#include <string>

namespace anna {
namespace xml {
class Node;
}
}


namespace anna {

namespace statistics {


class Operation {
public:

  enum Type {
    //Sum = 0, // is not very useful
    Average,
    StandardDeviation,
    BesselStandardDeviation, //This correction (the use of N - 1 instead of N) is known as Bessel's correction
    Minimum,
    Maximum
  };
};

typedef struct {

  unsigned long long Size;                           // Sample size
  double Sum;                                        // Accumulated sum
  double SquareSum;                                  // Accumulated sum of square values
  double Average;                                    // Current average
  //double SquareSumSampleDeviation;                   // Current total sum for square deviation items;
  double MinimumProcessed;                           // Minimum value processed
  double MaximumProcessed;                           // Maximum value processed
  anna::Millisecond MinimumEventTimestampMs;       // Timestamp (ms) for minimum event
  anna::Millisecond MaximumEventTimestampMs;       // Timestamp (ms) for maximum event
  anna::Millisecond LastResetEventTimestampMs;     // Timestamp (ms) for last statistics reset


  void reset() {
    Size = (unsigned long long)0;
    Sum = (double)0;
    SquareSum = (double)0;
    Average = (double)0;
    //SquareSumSampleDeviation = (double)0;
    MinimumProcessed = (double)0;
    MaximumProcessed = (double)0;
    MinimumEventTimestampMs = (anna::Millisecond)0;
    MaximumEventTimestampMs = (anna::Millisecond)0;
    LastResetEventTimestampMs = anna::Millisecond::getTime();
  }

} _concept_data_t;

typedef std::map <int, _concept_data_t> _concept_data_map_t;
typedef std::map <int, _concept_data_t>::const_iterator _concept_data_map_iter;




//------------------------------------------------------------------------------
//------------------------------------------------------------ class Accumulator
//------------------------------------------------------------------------------
/**
 * Accumulated data information for statistic sample.
 * It can manage a set of concepts at the same time.
 *
 * @short Contains statistical sample information
 */
class Accumulator {

  std::string a_name;

  void initialize(const int & conceptId) throw();
  _concept_data_t * getConcept(const int & conceptId) const throw(anna::RuntimeException);
  // Gets the data structure for a existing concept Id.
  // Initializes and returns the new data structure for a new stored concept Id which is valid (exists for the engine).
  // Launch exception if concept id is not a valid concept registered at Engine.

  /*mutable */_concept_data_map_t a_concept_data_map;

  std::string floatFormat(const int & numberOfDecimals) const throw();

  double getStandardDeviation(const _concept_data_t * conceptData) const throw(anna::RuntimeException);
  double getBesselStandardDeviation(const _concept_data_t * conceptData) const throw(anna::RuntimeException);


public:

  /**
  * Constructor.
  *
  * @param name Accumulator name
  */
  Accumulator(const std::string &name) : a_name(name) {};

  /**
  * Destructor.
  */
  ~Accumulator() {;}


  /**
   * Sets the accumulator name
   *
   * @param name Name provided
   */
  void setName(const std::string &name) throw() { a_name = name; }

  /**
  * Adds a new statistic concept through the accumulator, to ease the concept name creation,
  * which will be a string defined by: concept name + ': ' + accumulator name.
  *
  * @param description Concept description; e.g.: processing time, messages size, etc.
  * @param unit Concept unit description
  * @param integerNatureSample Most of cases we will measure 'time' with the unit which force integer values
  *                            (is more intuitive 850 msecs than 0,850 secs). Then, it is @em true by default.
  *                            This is useful to advice better representation for some indicators like minimum/maximum
  *                             within integer samples.
  * @param conceptNameFormat Format to register the complete concept name. Will be parsed using (in order) the provided
  *                          description, and the accumulator name: '<concept description> [<accumulator name>]' by default.
  *
  * @return Assigned concept identification number (sequence)
  */
  int addConcept(const std::string & description, const std::string & unit, const bool & integerNatureSample = true, const char *conceptNameFormat = "%s [%s]") throw();

  /**
  * Process new value for the sample regarding a concept identifier previously registered through the engine.
  *
  * @param conceptId statistical concept processed
  * @param value Value for processed item
  *
  * @see reset()
  */
  void process(const int & conceptId, const double & value) throw(anna::RuntimeException);


  /**
  * Reset sample for all concepts
  *
  * @see process()
  */
  void reset(void) throw();


  /**
  * Reset sample for provided concept
  *
  * @param conceptId Concept identification to be reset
  *
  * @see process()
  */
  void reset(const int & conceptId) throw(anna::RuntimeException);


  /**
  * Copy operator
  *
  * @param accumulator Class instance source of data
  *
  * @return Returns copied reference
  */
  const Accumulator & operator = (const Accumulator & accumulator);



  // Gets

  /**
   * Gets the accumulator name
   *
   * @result Accumulator name
   */
  const std::string &getName() const throw() { return a_name; }

  /**
  * Gets current sample size for any concept id
  *
  * @param conceptId Concept identification
  *
  * @return Sample size
  *
  * @see getValue()
  * @see asXML()
  */
  unsigned long long int sampleSize(const int & conceptId) const throw(anna::RuntimeException);


  /**
  * Get statistical information
  *
  * @param conceptId Statistical concept
  * @param operation Solicited operation
  *
  * @return Value for solicited conceptId/operation
  *
  * @see sampleSize()
  * @see asXML()
  */
  double getValue(const int & conceptId, const Operation::Type & operation) const throw(anna::RuntimeException);


  /**
  * Class string representation
  *
  * @param numberOfDecimals Number of float decimals at representation string. Default is 2.
  *
  * @return String with class content
  */
  std::string asString(const int & numberOfDecimals = 2) const throw();


  /**
  * Class XML representation
  *
  * @param numberOfDecimals Number of float decimals at XML representation. Default is 2.
  *
  * @return XML with class content
  */
  anna::xml::Node* asXML(anna::xml::Node* parent, const int & numberOfDecimals = 2) const throw();
};

}
}

#endif
