// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/functions.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/ServerSocket.hpp>
#include <anna/comm/internal/BinderSocket.hpp>
#include <anna/comm/ClientSocket.hpp>

using namespace std;
using namespace anna;

//static
const Millisecond comm::ServerSocket::DefaultBindDelay(200);

comm::ServerSocket::~ServerSocket() {
   delete a_binderSocket;
}

void comm::ServerSocket::prepare()
throw(RuntimeException) {
   if (::listen(Socket::a_fd, a_backlog) == -1) {
      const int xerrno(errno);
      std::string msg(asString());
      msg += " | listen";
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
   }
}

int comm::ServerSocket::do_bind(const struct sockaddr* s, const int len)
throw(RuntimeException) {
   if (a_sharedBind == false)
      return Socket::do_bind(s, len);

   if (a_binderSocket == NULL)
      a_binderSocket = new BinderSocket(this);

   a_binderSocket->requestBind(s, len);
   return 0;
}
/*
 * (1) Se invoca desde [Tx], pero hay que evitar que el .create se invoque a la misma vez que algn [Tz] este invocando
 * al ServerSocket::release porque ha cerrado la conexin.
 */
comm::LocalConnection* comm::ServerSocket::accept()
throw(RuntimeException) {
   LOGMETHOD(TraceMethod traceMethod(Logger::Local7, "comm::ServerSocket", "accept", ANNA_FILE_LOCATION));
   LocalConnection* result(NULL);
   sockaddr_in sourceAddress;
   socklen_t len(sizeof(sockaddr_in));
   anna_memset(&sourceAddress, 0, sizeof(sockaddr_in));
   int newSocket = ::accept(Socket::a_fd, (sockaddr*) & sourceAddress, &len);

   if (newSocket < 0) {
      const int xerrno = errno;

      if (xerrno != EWOULDBLOCK && xerrno != EINTR)
         throw RuntimeException(asString(), xerrno, ANNA_FILE_LOCATION);
   } else {
      result = a_localConnections.create();              // (1)

      try {
         result->setServerSocket(this);
         ClientSocket* clientSocket;

         if ((clientSocket = result->getClientSocket()) == NULL) {
            clientSocket = allocateClientSocket();
            result->setClientSocket(clientSocket);
            LOGDEBUG(
               string msg("comm::ServerSocket::accept | New ClientSocket: ");
               msg += functions::asHexString(anna_ptrnumber_cast(clientSocket));
               msg += functions::asText(" | fd: ", newSocket);
               Logger::debug(msg, ANNA_FILE_LOCATION);
            );
         } else {
            LOGDEBUG(
               string msg("comm::ServerSocket::accept | Reuse ClientSocket: ");
               msg += functions::asHexString(anna_ptrnumber_cast(clientSocket));
               msg += functions::asText(" | fd: ", newSocket);
               Logger::debug(msg, ANNA_FILE_LOCATION);
            );
         }

         clientSocket->setfd(newSocket);
         clientSocket->setCategory(getCategory());
         clientSocket->setReceiverFactory(*getReceiverFactory());
         LOGDEBUG(
            string msg("comm::ServerSocket::accept | ");
            msg += asString();
            msg += " | ";
            msg += result->asString();
            Logger::debug(msg, ANNA_FILE_LOCATION);
         );
      } catch (RuntimeException&) {
         ::close(newSocket);
         release(result);
         throw;
      }
   }

   return result;
}

comm::ClientSocket* comm::ServerSocket::allocateClientSocket() const
throw() {
   return new ClientSocket(getTransportFactory(), Socket::Domain::Inet, Socket::Type::Stream);
}

/*
 * Se invoca desde el [Tz] = LocalConnection:[Tx] -> Communicator
 * De forma que hay que bloquear esta instancia para que SU [Tx] no vaya a crear en este momento una nueva conexin.
 *
 *
 * Se invoca desde comm::handler::ServerSocket::apply::[Tx] -> <null>
 */
void comm::ServerSocket::release(LocalConnection* localConnection)
throw(RuntimeException) {
   if (localConnection == NULL)
      return;

   if (localConnection != NULL)
      localConnection->setServerSocket(NULL);

   a_localConnections.release(localConnection);
}

std::string comm::ServerSocket::asString() const
throw() {
   std::string msg("comm::ServerSocket { ");
   msg += comm::Socket::asString();
   msg += " | Bind: ";

   if (a_sharedBind == true) {
      msg += "Shared | ";
      msg += functions::asString(a_binderSocket);
   } else
      msg += "Exclusive";

   return msg += " }";
}

xml::Node* comm::ServerSocket::asXML(xml::Node* parent) const
throw(RuntimeException) {
   xml::Node* result = parent->createChild("comm.ServerSocket");
   comm::Socket::asXML(result);
   result->createAttribute("Bind", (a_sharedBind) ? "Shared" : "Exclusive");

   if (a_sharedBind == true && a_binderSocket != NULL)
      a_binderSocket->asXML(result);

   return result;
}
