// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/codec/basetypes/Address.hpp>

// Standard
#include <arpa/inet.h>

#include <anna/core/functions.hpp>


//------------------------------------------------------------------------------
//------------------------------------------------------- Address::updateBasic()
//------------------------------------------------------------------------------
void anna::diameter::codec::basetypes::Address::updateBasic() throw(anna::RuntimeException) {
  std::string result;
  // address version
  result.push_back((S8)(a_address.Version >> 8));
  result.push_back((S8)a_address.Version);

  // address value
  if(a_address.isIPv4()) {
    if(a_abbreviatePresentation) {
      unsigned char buf[sizeof(struct in6_addr)];
      int s = inet_pton(AF_INET, a_address.Value.c_str(), buf);

      if(s > 0) {
        result.push_back((S8) buf[0]);
        result.push_back((S8) buf[1]);
        result.push_back((S8) buf[2]);
        result.push_back((S8) buf[3]);
      } else {
        if(s < 0) perror("inet_pton");

        throw anna::RuntimeException("Address::setValue | Wrong IPv4 address format", ANNA_FILE_LOCATION);
      }
    } else {
      if(!anna::functions::isIPv4(a_address.Value))
        throw anna::RuntimeException("Address::setValue | Wrong IPv4 address format", ANNA_FILE_LOCATION);

      int dec1, dec2, dec3, dec4; // U32 type is valid, 'int' better: argument for %d must be 'int *' (http://www.cplusplus.com/reference/clibrary/cstdio/sscanf/)
      sscanf(a_address.Value.c_str(), "%d.%d.%d.%d", &dec1, &dec2, &dec3, &dec4);
      result.push_back((S8) dec1);
      result.push_back((S8) dec2);
      result.push_back((S8) dec3);
      result.push_back((S8) dec4);
    }
  } else if(a_address.isIPv6()) {
    if(a_abbreviatePresentation) {
      unsigned char buf[sizeof(struct in6_addr)];
      int s = inet_pton(AF_INET6, a_address.Value.c_str(), buf);

      if(s > 0) {
        result.push_back((S8) buf[0]);
        result.push_back((S8) buf[1]);
        result.push_back((S8) buf[2]);
        result.push_back((S8) buf[3]);
        result.push_back((S8) buf[4]);
        result.push_back((S8) buf[5]);
        result.push_back((S8) buf[6]);
        result.push_back((S8) buf[7]);
        result.push_back((S8) buf[8]);
        result.push_back((S8) buf[9]);
        result.push_back((S8) buf[10]);
        result.push_back((S8) buf[11]);
        result.push_back((S8) buf[12]);
        result.push_back((S8) buf[13]);
        result.push_back((S8) buf[14]);
        result.push_back((S8) buf[15]);
      } else {
        if(s < 0) perror("inet_pton");

        throw anna::RuntimeException("Address::setValue | Wrong IPv6 address format", ANNA_FILE_LOCATION);
      }
    } else {
      if(!anna::functions::isIPv6(a_address.Value))
        throw anna::RuntimeException("Address::setValue | Wrong IPv6 address format", ANNA_FILE_LOCATION);

      int hex1, hex2, hex3, hex4, hex5, hex6, hex7, hex8; // U16 type is not valid: argument for %X must be 'int *' (http://www.cplusplus.com/reference/clibrary/cstdio/sscanf/)
      sscanf(a_address.Value.c_str(), "%X:%X:%X:%X:%X:%X:%X:%X", &hex1, &hex2, &hex3, &hex4, &hex5, &hex6, &hex7, &hex8);
      result.push_back((S8)(hex1 >> 8));
      result.push_back((S8)(hex1 & 0x00FF));
      result.push_back((S8)(hex2 >> 8));
      result.push_back((S8)(hex2 & 0x00FF));
      result.push_back((S8)(hex3 >> 8));
      result.push_back((S8)(hex3 & 0x00FF));
      result.push_back((S8)(hex4 >> 8));
      result.push_back((S8)(hex4 & 0x00FF));
      result.push_back((S8)(hex5 >> 8));
      result.push_back((S8)(hex5 & 0x00FF));
      result.push_back((S8)(hex6 >> 8));
      result.push_back((S8)(hex6 & 0x00FF));
      result.push_back((S8)(hex7 >> 8));
      result.push_back((S8)(hex7 & 0x00FF));
      result.push_back((S8)(hex8 >> 8));
      result.push_back((S8)(hex8 & 0x00FF));
    }
  } else if(a_address.isE164()) {
    result += a_address.Value; // direct assignment
  } else {
    result += a_address.Value; // direct assignment
  }

  OctetString::setValue(result);
}


//------------------------------------------------------------------------------
//------------------------------------------------ Address::setPrintableString()
//------------------------------------------------------------------------------
void anna::diameter::codec::basetypes::Address::setPrintableString(const char * printableString) throw(anna::RuntimeException) {
  // First: User will update child class members
  // Auxiliary
  iana_address_t address;
  std::string value;
  // From message.dtd:
  // [ Address ] ('<type (IANA Address Family Number)>|<value>' representation; i.e. '1|192.168.0.1'(IPv4), '8|34616279266'(E164), etc.
  //             Type (and pipe) field could be avoided for IPv4 and IPv6 address types (a light parse checking is done: one colon for
  //             IPv6, one dot for IPv4). Internal engine always includes type on data field, which is also recommended for inputs.
  //             Currently, only IPv4, IPv6 and E164 address types have a known printable presentation)
  std::string source = printableString;
  size_t pipePos = source.find("|");

  if(pipePos != std::string::npos) {
    if(pipePos == 0)
      throw anna::RuntimeException("Address::fromPrintableString | Unreconized printable source: missing type before pipe (see 'Diameter message DTD' for [ Address ] printable format)", ANNA_FILE_LOCATION);

    std::string beforePipe(printableString, pipePos);
    address.Version = atoi(beforePipe.c_str());
    address.Value.assign(printableString, pipePos + 1, source.size() - pipePos - 1);
  }
  // For backward compatibility (configurations, tests cases, etc.)
  else { // only IPv4/v6 addresses are supported with a light checking (version pre-assumption): dot for IPv4, colon for IPv6. Anyway, updateBasic() will do regexp for supposed version
    if(strstr(printableString, ":"))
      address.setIPv6(printableString);
    else if(strstr(printableString, "."))
      address.setIPv4(printableString);
    else
      throw anna::RuntimeException("Address::fromPrintableString | Unreconized printable source (see 'Diameter message DTD' for [ Address ] printable format)", ANNA_FILE_LOCATION);
  }

  // Assignments
  setIANAAddress(address);
}


//------------------------------------------------------------------------------
//------------------------------------------------ Address::setPrintableString()
//------------------------------------------------------------------------------
std::string anna::diameter::codec::basetypes::Address::asPrintableString() throw(anna::RuntimeException) {
  bool knownPrintablePresentation = (a_address.isIPv4() || a_address.isIPv6() || a_address.isE164());

  if(!knownPrintablePresentation)
    throw anna::RuntimeException("Address::asPrintableString | Only IPv4, IPv6 and E164 have a known printable presentation", ANNA_FILE_LOCATION);

  return anna::functions::asString("%d|%s", a_address.getVersion(), a_address.getValue());
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Address::decode()
//------------------------------------------------------------------------------
void anna::diameter::codec::basetypes::Address::decode(const char* buffer, const int size) throw(anna::RuntimeException) {
  if(!buffer)
    throw anna::RuntimeException("Address::decode | Null Buffer provided", ANNA_FILE_LOCATION);

  char str[INET6_ADDRSTRLEN];

  // IP Version
  if(size < 2)
    throw anna::RuntimeException("Address::decode | At least 2 bytes needed to decode address version", ANNA_FILE_LOCATION);

  a_address.Version = (((U16)buffer[0] << 8) & 0xFF00) +
                      ((U16)buffer[1]        & 0x00FF); // two first octets

  // IP Address
  if(a_address.isIPv4()) {  // 6 bytes
    if(size != 6)
      throw anna::RuntimeException("Address::decode | 2+4(=6) bytes needed to decode IPv4 address", ANNA_FILE_LOCATION);

    if(a_abbreviatePresentation) {
      if(inet_ntop(AF_INET, buffer + 2, str, INET_ADDRSTRLEN) != NULL) {
        a_address.Value = str;
      } else {
        throw anna::RuntimeException("Address::decode | Error decoding IPv4 address", ANNA_FILE_LOCATION);
      }
    } else {
      a_address.Value = anna::functions::asString("%d.%d.%d.%d", (U8)buffer[2], (U8)buffer[3], (U8)buffer[4], (U8)buffer[5]);
    }
  } else if(a_address.isIPv6()) {  // 18 bytes
    if(size != 18)
      throw anna::RuntimeException("Address::decode | 2+16(=18) bytes needed to decode IPv6 address", ANNA_FILE_LOCATION);

    if(a_abbreviatePresentation) {
      if(inet_ntop(AF_INET6, buffer + 2, str, INET6_ADDRSTRLEN) != NULL) {
        a_address.Value = str;
      } else {
        throw anna::RuntimeException("Address::decode | Error decoding IPv6 address", ANNA_FILE_LOCATION);
      }
    } else {
      a_address.Value = anna::functions::asString("%X:%X:%X:%X:%X:%X:%X:%X",
                        ((((U8)buffer[2]) << 8) & 0xFF00 /* same as (U16 cast)*/) +
                        (((U8)buffer[3])        & 0x00FF),
                        ((((U8)buffer[4]) << 8) & 0xFF00) +
                        (((U8)buffer[5])        & 0x00FF),
                        ((((U8)buffer[6]) << 8) & 0xFF00) +
                        (((U8)buffer[7])        & 0x00FF),
                        ((((U8)buffer[8]) << 8) & 0xFF00) +
                        (((U8)buffer[9])        & 0x00FF),
                        ((((U8)buffer[10]) << 8) & 0xFF00) +
                        (((U8)buffer[11])        & 0x00FF),
                        ((((U8)buffer[12]) << 8) & 0xFF00) +
                        (((U8)buffer[13])        & 0x00FF),
                        ((((U8)buffer[14]) << 8) & 0xFF00) +
                        (((U8)buffer[15])        & 0x00FF),
                        ((((U8)buffer[16]) << 8) & 0xFF00) +
                        (((U8)buffer[17])        & 0x00FF));
    }
  } else if(a_address.isE164()) {  // n bytes = 2 + address length
    a_address.Value.assign(buffer + 2, size - 2);  // Direct assignment
  } else { // Generic decoding (could support non printable addresses but they must be...)
    a_address.Value.assign(buffer + 2, size - 2);
  }

  // Base class decode()
  OctetString::decode(buffer, size);
}
