// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Standard
#include <string>
#include <iostream>
#include <errno.h>

#include <signal.h>	// sigaction, sigemptyset, struct sigaction, SIGCHLD, SA_RESTART, SA_NOCLDSTOP
#include <stdio.h>	// perror
#include <stdlib.h>	// exit
#include <sys/wait.h>	// waitpid, pid_t, WNOHANG
#include <climits>

// cmd with fork:
#include <sys/types.h>
#include <unistd.h>


// Project
#include <anna/testing/TestStep.hpp>

#include <anna/testing/defines.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter.comm/OriginHost.hpp>
#include <anna/testing/TestCase.hpp>
#include <anna/testing/TestManager.hpp>
#include <anna/testing/TestTimer.hpp>
#include <anna/xml/Compiler.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/LocalServer.hpp>
#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/ServerSession.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/helpers/base/functions.hpp>



using namespace anna::testing;


namespace {

  void handle_sigchld(int sig) {
    while (waitpid((pid_t)(-1 /* any child (the only) */), 0, WNOHANG|WNOWAIT) > 0) {}
  }

  void cmdRunOnThread (TestStepCmd *step, const std::string &cmd) {

    // Thread running:
    step->setThreadRunning(true);

    int status = -2;

    struct sigaction sa;
    sa.sa_handler = &handle_sigchld;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART | SA_NOCLDSTOP;
    if (sigaction(SIGCHLD, &sa, 0) != -1) {

      status = system(cmd.c_str());
      /* POPEN version:
          char readbuf[256];
          FILE *fp = popen(cmd.c_str(), "r");
          if (fp) {
            while(fgets(readbuf, sizeof(readbuf), fp))
              step->appendOutput("\n");
              step->appendOutput(readbuf);
              status = pclose(fp);
          }
          else {
            status = -1;
          }
       */
    }
    else {
      perror(0);
    }
    // This can be implemented portably and somewhat more concisely with the signal function if you prefer:
    // if (signal(SIGCHLD, SIG_IGN) == SIG_ERR) {
    //   perror(0);
    //   exit(1);
    // }

    if (status < 0) {
      char buf[256];
      char const * str = strerror_r(errno, buf, 256);
      step->setErrorMsg(anna::functions::asString("errno = %d (%s)", errno, str));
    }

    step->setResultCode(WEXITSTATUS(status)); // rc = status >>= 8; // divide by 256
    step->complete();
    // TODO: terminate thread when deprecated (RT signal ?)
    // TODO: mutex the step while setting data here !!
  }

/*
  void cmdRunOnThreadWithFork (TestStepCmd *step, const std::string &cmd) {

    // Thread running:
    step->setThreadRunning(true);

    pid_t cpid, w;
    int status = -2;

    if ((cpid = fork()) < 0) {
      step->setErrorMsg("Error in fork()");
    }
    else if (cpid == 0) {
      // child
      status = system(cmd.c_str());
      _exit(WEXITSTATUS(status));
    }
    else {
      // parent
      step->setChildPid(cpid);
      do {
        w = waitpid(cpid, &status, WUNTRACED | WCONTINUED);
        if (w != -1) {

          if (WIFEXITED(status)) {
            step->setResultCode(WEXITSTATUS(status)); // rc = status >>= 8; // divide by 256
            break;
          }
          else if (WIFSIGNALED(status)) {
            step->setErrorMsg(anna::functions::asString("killed by signal %d", WTERMSIG(status)));
            step->setResultCode(128 + WTERMSIG(status));
            break;
          } else if (WIFSTOPPED(status)) {
            step->setErrorMsg(anna::functions::asString("stopped by signal %d", WSTOPSIG(status)));
          } else if (WIFCONTINUED(status)) {
            step->setErrorMsg("continued");
          }
        }
        else {
          step->setErrorMsg("waitpid error");
          step->setResultCode(-1);
          break;
        }
      } while (!WIFEXITED(status) && !WIFSIGNALED(status));

      step->complete();
    }
  }
*/

  bool decodeMessage(const anna::DataBlock &message, anna::diameter::codec::Message &messageCodec) throw() {

    if (message.isEmpty())
      return false;

    bool result = true;
    try {
      messageCodec.clear();
      messageCodec.decode(message);
    }
    catch (anna::RuntimeException &ex) {
      ex.trace();
      result = false;
    }

    return result;
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStep
////////////////////////////////////////////////////////////////////////////////////////////////////////
void TestStep::initialize(TestCase *testCase) {
  a_testCase = testCase;
  a_completed = false;
  a_type = Type::Unconfigured;
  a_beginTimestamp = 0;
  a_endTimestamp = 0;
  a_number = testCase->steps() + 1; // testCase is not NULL
}

bool TestStep::decodeMessage(bool trust) throw() {
  if (a_messageCodec) return true;
  a_messageCodec = new anna::diameter::codec::Message;
  if (::decodeMessage(a_message, *a_messageCodec)) return true;
  if (trust) {
    LOGDEBUG(anna::Logger::debug("Error DECODING, but trusting it ...", ANNA_FILE_LOCATION));
    return true;
  }

  delete a_messageCodec;
  a_messageCodec = NULL;
  return false;
}

const char* TestStep::asText(const Type::_v type)
throw() {
  static const char* text [] = { "Unconfigured", "Timeout", "SendDiameterXmlToEntity", "SendDiameterXmlToClient", "Delay", "Wait", "Command", "IpLimit" };
  return text [type];
}

anna::xml::Node* TestStep::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = parent->createChild("TestStep");

  result->createAttribute("Number", a_number);
  result->createAttribute("Type", asText(a_type));
  result->createAttribute("Completed", (a_completed ? "yes":"no"));

  // Begin
  std::string s_aux = a_beginTimestamp.asString();
  //  int deltaMs = (int)(a_beginTimestamp - a_testCase->getStartTimestamp());
  //  if (a_beginTimestamp != 0 && deltaMs > 0) s_aux += anna::functions::asString(" [%.3f]", deltaMs/1000.0);
  result->createAttribute("BeginTimestamp", s_aux);

  // End
  s_aux = a_endTimestamp.asString();
  //  deltaMs = (int)(a_endTimestamp - a_testCase->getStartTimestamp());
  //  if (a_endTimestamp != 0 && deltaMs > 0) s_aux += anna::functions::asString(" [%.3f]", deltaMs/1000.0);
  result->createAttribute("EndTimestamp", s_aux);

  if (a_beginTimestamp != 0 && a_endTimestamp != 0)
    result->createAttribute("LapseMilliseconds", getLapseMs());

  return result;
}

std::string TestStep::asXMLString() throw() {
  anna::xml::Node root("root");
  return anna::xml::Compiler().apply(asXML(&root));
}

bool TestStep::execute() throw() {

  int ia = a_testCase->interactiveAmount();
  if (ia > -1) {
    if (ia == 0) return false;
    a_testCase->interactiveExecution();
    LOGDEBUG(anna::Logger::debug("Interactive execution ...", ANNA_FILE_LOCATION));
    if (a_executed) return false; // avoid repeating (this implies amount consumption)
  }

  const char *literal = "Execute %s Test Step %d/%d (test case %llu)";

  TestManager& testManager (TestManager::instantiate ());
  if (testManager.getDumpStdout()) {
    bool dump = (a_type == Type::Wait) ? (!a_completed) : true;
    if (dump) {
    std::cout << std::endl << "=> " << anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()) << std::endl;
    }
  }

  LOGDEBUG(anna::Logger::debug(anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()), ANNA_FILE_LOCATION));
  setBeginTimestamp(anna::functions::millisecond());
  a_executed = true;
  return do_execute();
}

void TestStep::complete() throw() {

  const char *literal = "Complete %s Test Step %d/%d (test case %llu)";

  TestManager& testManager (TestManager::instantiate ());
  if (testManager.getDumpStdout()) {
    std::cout << "=> " << anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()) << std::endl;
  }

  LOGDEBUG(anna::Logger::debug(anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()), ANNA_FILE_LOCATION));
  a_completed = true;
  setEndTimestamp(anna::functions::millisecond());
  do_complete();
}

void TestStep::reset() throw() {

  const char *literal = "Reset %s Test Step %d/%d (test case %llu)";

  TestManager& testManager (TestManager::instantiate ());
  if (testManager.getDumpStdout()) {
    std::cout << "=> " << anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()) << std::endl;
  }

  LOGDEBUG(anna::Logger::debug(anna::functions::asString(literal, asText(a_type), a_number, a_testCase->steps(), a_testCase->getId()), ANNA_FILE_LOCATION));
  // type and testCase kept
  a_completed = false;
  a_executed = false;
  a_beginTimestamp = 0;
  a_endTimestamp = 0;
  do_reset();
}

void TestStep::next() throw() {
  a_testCase->nextStep();
  a_testCase->process();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStepTimeout
////////////////////////////////////////////////////////////////////////////////////////////////////////
anna::xml::Node* TestStepTimeout::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent); // end timestamp will be 0 if test finished OK
  //parent->createChild("TestStepTimeout");
  result->createAttribute("Timeout", a_timeout.asString());

  return result;
}

bool TestStepTimeout::do_execute() throw() {
  try {
    a_timer = TestManager::instantiate().createTimer((TestCaseStep*)this, a_timeout, TestTimer::Type::TimeLeft);
  }
  catch (anna::RuntimeException &ex) {
    ex.trace();
    a_testCase->addDebugSummaryHint(ex.asString()); // before report (when set Failed state)
    a_testCase->setState(TestCase::State::Failed);
  }

  return true; // go next
}

void TestStepTimeout::do_complete() throw() {
  int stepNumber = getNumber();
  if (stepNumber == a_testCase->steps()) {
    a_testCase->addDebugSummaryHint(anna::functions::asString("Timeout expired (step number %d) but it was the last test case step", stepNumber)); // before report (when set Failed state)
    a_testCase->setState(TestCase::State::Success);
    if (TestManager::instantiate().getDumpStdout()) {
      std::cout << "Expired Timeout Test LAST Step: considered SUCCESSFUL TEST CASE" << std::endl;
    }
  }
  else if (a_testCase->getState() == TestCase::State::InProgress) { // sure
    a_testCase->addDebugSummaryHint(anna::functions::asString("Timeout expired (step number %d) before test case finished", stepNumber)); // before report (when set Failed state)
    a_testCase->setState(TestCase::State::Failed);
    if (TestManager::instantiate().getDumpStdout()) {
      std::cout << "Expired Timeout Test Step: FAILED TEST CASE" << std::endl;
    }
  }

  a_timer = NULL;
}

void TestStepTimeout::cancelTimer() throw() {
  if (!a_timer) return;
  try {
    TestManager::instantiate().cancelTimer(a_timer);
  }
  catch (anna::RuntimeException &ex) {
    ex.trace();
  }
  a_timer = NULL;
  //a_timeout = 0; THIS IS CONFIGURATION INFO
}

void TestStepTimeout::do_reset() throw() {
  cancelTimer();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStepSendDiameterXml
////////////////////////////////////////////////////////////////////////////////////////////////////////
anna::xml::Node* TestStepSendDiameterXml::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent);
  //parent->createChild("TestStepSendDiameterXml");
  std::string msg = "", xmlmsg = "";

  // Message
  if (TestManager::instantiate().getDumpHex()) {
    if (a_message.isEmpty()) {
      msg = "<empty>";
    }
    else {
      msg = "\n"; msg += a_message.asString(); msg += "\n";
    }
  }

  if (decodeMessage()) {
    xmlmsg = "\n";
    xmlmsg += a_messageCodec->asXMLString();
    xmlmsg += "\n";
  }
  else {
    xmlmsg = "<unable to decode, check traces>";
  }

  if (msg != "") result->createAttribute("Message", msg);
  if (xmlmsg != "") result->createAttribute("XMLMessage", xmlmsg);
  result->createAttribute("Expired", (a_expired ? "yes":"no"));
  if (a_waitForRequestStepNumber != -1)
    result->createAttribute("WaitForRequestStepNumber", a_waitForRequestStepNumber);

  return result;
}

bool TestStepSendDiameterXml::do_execute() throw() {
  bool success = false;
  std::string failReason = "Error sending diameter message";
  anna::diameter::comm::Entity *entity = a_originHost->getEntity(); // by default
  anna::diameter::comm::LocalServer *localServer = a_originHost->getDiameterServer(); // by default
  const TestStepWaitDiameter *tsw = NULL;
  anna::diameter::comm::Message *msg = a_originHost->createCommMessage();

  try {

    if (a_waitForRequestStepNumber != -1) {
      bool thisIsAnswer = anna::diameter::codec::functions::isAnswer(getMsgDataBlock());
      LOGDEBUG(
         std::string trace = anna::functions::asString("'Wait For Request' step number for this %s: %d", (thisIsAnswer ? "answer":"request"), a_waitForRequestStepNumber);
         anna::Logger::debug(trace, ANNA_FILE_LOCATION);
      );

      // Referenced request in the 'wait for request step':
      tsw = static_cast<const TestStepWaitDiameter*>(a_testCase->getStep(a_waitForRequestStepNumber));
      const anna::DataBlock &referenceRequest = tsw->getMsgDataBlock();
      std::string sessionIdReferenceRequest = anna::diameter::helpers::base::functions::getSessionId(referenceRequest);

      if (thisIsAnswer) { // is an answer: try to copy sequence information; alert about Session-Id discrepance
        anna::diameter::HopByHop hbh = anna::diameter::codec::functions::getHopByHop(referenceRequest);
        anna::diameter::EndToEnd ete = anna::diameter::codec::functions::getEndToEnd(referenceRequest);

        // Update sequence:
        anna::diameter::codec::functions::setHopByHop(a_message, hbh);
        anna::diameter::codec::functions::setEndToEnd(a_message, ete);
      }

      // Session-Id substitution:
      std::string thisSessionId = anna::diameter::helpers::base::functions::getSessionId(getMsgDataBlock());
      if (thisSessionId != sessionIdReferenceRequest) {
        static anna::diameter::codec::Message codecMsg;
        codecMsg.decode(getMsgDataBlock());
        codecMsg.getAvp("Session-Id")->getUTF8String()->setValue(sessionIdReferenceRequest);
        a_message = codecMsg.code();
        std::string trace = anna::functions::asString("Replacing %s Session-Id (%s) to set the corresponding request one (%s)", (thisIsAnswer ? "answer":"request"), thisSessionId.c_str(), sessionIdReferenceRequest.c_str());
        LOGDEBUG(anna::Logger::debug(trace, ANNA_FILE_LOCATION));
        a_testCase->addDebugSummaryHint(trace);
      }
    }

    // Create comm message:
    //msg->clearBody();
    msg->setBody(a_message);


    if (getType() == Type::Sendxml2e) {
      anna::diameter::comm::ClientSession *usedClientSession = NULL;

      if (tsw) { // is an answer for a received request on wait condition
        anna::diameter::comm::ClientSession *clientSession = tsw->getClientSession();
        if (clientSession) {
          /* response NULL (is an answer) */clientSession->send(msg);
          success = true;
          usedClientSession = clientSession;
        }
        else {
          failReason = "Reference wait step didn't store a valid client session. Unable to send the message";
          LOGWARNING(anna::Logger::warning(failReason, ANNA_FILE_LOCATION));
        }
      }
      else {
        if (entity) {
          success = entity->send(msg);
          anna::diameter::comm::Server *usedServer = entity->getLastUsedResource();
          usedClientSession = usedServer ? usedServer->getLastUsedResource() : NULL;
        }
        else {
          failReason = "There is no diameter entity currently configured. Unable to send the message";
          LOGWARNING(anna::Logger::warning(failReason, ANNA_FILE_LOCATION));
        }
      } // else (normal sending)

      // Detailed log:
      if(a_originHost->logEnabled()) {
        if (decodeMessage(true /* trust */)) {
          std::string detail = usedClientSession ? usedClientSession->asString() : "<null client session>"; // shouldn't happen
          a_originHost->writeLogFile(*a_messageCodec, (success ? "sent2e" : "send2eError"), detail);
        }
      }
    }
    else if (getType() == Type::Sendxml2c) {
      anna::diameter::comm::ServerSession *usedServerSession = NULL;

      if (tsw) { // is an answer for a received request on wait condition
        anna::diameter::comm::ServerSession *serverSession = tsw->getServerSession();
        if (serverSession) {
          /* response NULL (is an answer) */serverSession->send(msg);
          success = true;
          usedServerSession = serverSession;
        }
        else {
          failReason = "Reference wait step didn't store a valid server session. Unable to send the message";
          LOGWARNING(anna::Logger::warning(failReason, ANNA_FILE_LOCATION));
        }
      }
      else {
        if (localServer) {
          success = localServer->send(msg);
          usedServerSession = localServer->getLastUsedResource();
        }
        else {
          failReason = "There is no diameter local server currently configured. Unable to send the message";
          LOGWARNING(anna::Logger::warning(failReason, ANNA_FILE_LOCATION));
        }
      } // else (normal sending)

      // Detailed log:
      if(a_originHost->logEnabled()) {
        if (decodeMessage(true /* trust */)) {
          std::string detail = usedServerSession ? usedServerSession->asString() : "<null server session>"; // shouldn't happen
          a_originHost->writeLogFile(*a_messageCodec, (success ? "sent2c" : "send2cError"), detail);
        }
      }
    }

  } catch(anna::RuntimeException &ex) {
    failReason = ex.asString();
  }

  // release msg
  a_originHost->releaseCommMessage(msg);

  if (!success) {
    a_testCase->addDebugSummaryHint(failReason); // before report (when set Failed state);
    a_testCase->setState(TestCase::State::Failed);
  }
  else {
    complete();
  }

  if (TestManager::instantiate().getDumpStdout()) {
    std::cout << "Executed SendDiameterXml Test Step: " << (success ? "OK":"ERROR") << std::endl;
  }

  return success; // go next if sent was OK
}

void TestStepSendDiameterXml::do_reset() throw() {
  a_expired = false;
  //a_message.clear();
  //a_messageAlreadyDecoded = false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStepDelay
////////////////////////////////////////////////////////////////////////////////////////////////////////
anna::xml::Node* TestStepDelay::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent);
  //parent->createChild("TestStepDelay");

  result->createAttribute("Delay", ((a_delay == 0) ? "dummy step, no delay" : a_delay.asString()));

  return result;
}

bool TestStepDelay::do_execute() throw() {
  if (a_delay == 0) { complete(); return true; } // special case
  try {
    a_timer = TestManager::instantiate().createTimer((TestCaseStep*)this, a_delay, TestTimer::Type::Delay);
  }
  catch (anna::RuntimeException &ex) {
    ex.trace();
    a_testCase->addDebugSummaryHint(ex.asString()); // before report (when set Failed state)
    a_testCase->setState(TestCase::State::Failed);
  }

  return false; // don't go next (wait complete)
}

void TestStepDelay::do_complete() throw() {
  if (a_delay == 0) return; // special case
  a_timer = NULL;
  if (TestManager::instantiate().getDumpStdout()) {
    std::cout << "Consumed Delay Test Step (" << a_delay << " milliseconds)" << std::endl;
  }
  next(); // next() invoked here because execute() is always false for delay and never advance the iterator
  // TODO, avoid this recursion
}

void TestStepDelay::cancelTimer() throw() {
  if (!a_timer) return;
  if (a_delay == 0) return; // special case
  try {
    TestManager::instantiate().cancelTimer(a_timer);
  }
  catch (anna::RuntimeException &ex) {
    ex.trace();
  }
  a_timer = NULL;
  //a_delay = 0; THIS IS CONFIGURATION INFO
}

void TestStepDelay::do_reset() throw() {
  cancelTimer();
}


////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStepWaitDiameter
////////////////////////////////////////////////////////////////////////////////////////////////////////
void TestStepWaitDiameter::setCondition(bool fromEntity,
    const std::string &code, const std::string &bitR, const std::string &hopByHop, const std::string &applicationId,
    const std::string &sessionId, const std::string &resultCode,
    const std::string &msisdn, const std::string &imsi, const std::string &serviceContextId) throw() {

  a_condition.setReceivedFromEntity(fromEntity);
  a_condition.setCode(code);
  a_condition.setBitR(bitR);
  a_condition.setResultCode(resultCode);
  a_condition.setSessionId(sessionId);
  a_condition.setHopByHop(hopByHop);
  a_condition.setMsisdn(msisdn);
  a_condition.setImsi(imsi);
  a_condition.setServiceContextId(serviceContextId);
}

void TestStepWaitDiameter::setConditionRegexpHex(bool fromEntity, const std::string &regexp) throw() {
  a_condition.setReceivedFromEntity(fromEntity);
  a_condition.setRegexpHex(regexp);
}

void TestStepWaitDiameter::setConditionRegexpXml(bool fromEntity, const std::string &regexp) throw() {
  a_condition.setReceivedFromEntity(fromEntity);
  a_condition.setRegexpXml(regexp);
}

anna::xml::Node* TestStepWaitDiameter::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent);
  //parent->createChild("TestStepWaitDiameter");
  std::string msg = "", xmlmsg = "";

  // Condition
  a_condition.asXML(result);

  // Message
  if (TestManager::instantiate().getDumpHex()) {
    if (a_message.isEmpty()) {
      msg = "<empty>";
    }
    else {
      msg = "\n"; msg += a_message.asString(); msg += "\n";
    }
  }

  if (a_message.isEmpty()) {
      xmlmsg = "<empty>";
  }
  else {
    if (decodeMessage()) {
      xmlmsg = "\n";
      xmlmsg += a_messageCodec->asXMLString();
      xmlmsg += "\n";
    }
    else {
      xmlmsg = "<unable to decode, check traces>";
    }
  }

  if (msg != "") result->createAttribute("MatchedMessage", msg);
  if (xmlmsg != "") result->createAttribute("MatchedXMLMessage", xmlmsg);
  if (a_clientSession) result->createAttribute("ClientSessionOrigin", anna::functions::asString("%p", a_clientSession));
  if (a_serverSession) result->createAttribute("ServerSessionOrigin", anna::functions::asString("%p", a_serverSession));

  return result;
}

bool TestStepWaitDiameter::do_execute() throw() {
  return a_completed;
}

void TestStepWaitDiameter::do_complete() throw() {
  //a_testCase->process(); // next() not invoked; we only want to reactivate the test case
  // avoid stack overflow: we will process the test case externally when incoming message is fulfilled (TestCase.cpp), and TestManager is noticed
}

bool TestStepWaitDiameter::fulfilled(const anna::DataBlock &db/*, bool matchSessionId*/) throw() {
  if (a_condition.comply(db/*, matchSessionId*/)) {
    a_message = db; // store matched
    complete();

    if (TestManager::instantiate().getDumpStdout()) {
      std::cout << "Fulfilled Wait diameter Test Step" << std::endl;
    }

    return true;
  }

  if (TestManager::instantiate().getDumpStdout()) {
    std::cout << "NOT Fulfilled Wait diameter Test Step" << std::endl;
  }

  return false;
}

void TestStepWaitDiameter::do_reset() throw() {
  a_message.clear();
  a_clientSession = NULL;
  a_serverSession = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////
// TestStepCmd
////////////////////////////////////////////////////////////////////////////////////////////////////////
anna::xml::Node* TestStepCmd::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent);
  //parent->createChild("TestStepCmd");

  result->createAttribute("Script", (a_script != "") ? a_script:"<no script>");
  if (a_errorMsg != "") result->createAttribute("ErrorMessage", a_errorMsg);
  if (a_threadRunning) {
    if (a_childPid != -1)
      result->createAttribute("ChildPid", a_childPid);
  }
  else {
    if (a_resultCode != -2) {
      result->createAttribute("ResultCode", a_resultCode);
      //if (a_output != "") result->createAttribute("Output", a_output);
    }
  }

  return result;
}

bool TestStepCmd::do_execute() throw() {
  if (!a_threadRunning /* || a_threadDeprecated DO NOT WANT TO OVERLAP ... */) {
    // Special tags to replace:
    std::string cmd = getScript();
    size_t index;
    while ((index = cmd.find(SH_COMMAND_TAG_FOR_REPLACE__CYCLE_ID)) != std::string::npos)
      cmd.replace(index, strlen(SH_COMMAND_TAG_FOR_REPLACE__CYCLE_ID), anna::functions::asString(TestManager::instantiate().getPoolCycle()));
    while ((index = cmd.find(SH_COMMAND_TAG_FOR_REPLACE__TESTCASE_ID)) != std::string::npos)
      cmd.replace(index, strlen(SH_COMMAND_TAG_FOR_REPLACE__TESTCASE_ID), anna::functions::asString(a_testCase->getId()));
    while ((index = cmd.find(SH_COMMAND_TAG_FOR_REPLACE__TESTSTEP_ID)) != std::string::npos)
      cmd.replace(index, strlen(SH_COMMAND_TAG_FOR_REPLACE__TESTSTEP_ID), anna::functions::asString(getNumber()));

    a_thread = std::thread(cmdRunOnThread, this, cmd);
    //a_thread = std::thread(cmdRunOnThreadWithFork, this, cmd);

    a_thread.detach();
  }

  return false; // don't go next (wait complete): If system function on thread stucks, then the reset test case will stuck here forever.
  // We must implement a interrupt procedure for the thread on reset call... TODO !
}

void TestStepCmd::do_complete() throw() {

  if (TestManager::instantiate().getDumpStdout()) {
    std::cout << "Executed Command Test Step (" << a_script << ") [rc=" << a_resultCode << "]" << std::endl;
  }

  a_threadRunning = false;
  if (a_threadDeprecated) {
    a_threadDeprecated = false;
    do_reset();
    setErrorMsg(anna::functions::asString("Step %d deprecated due to previous reset for Test Case %llu", getNumber(), a_testCase->getId()));
    a_testCase->setState(TestCase::State::Failed);
    return; // ignore TODO: interrupt the thread to avoid execution of the script
  }

  if (getResultCode() != 0)
    a_testCase->setState(TestCase::State::Failed);
  else
    next(); // next() invoked here because execute() is always false for delay and never advance the iterator
  // TODO, avoid this recursion
}

void TestStepCmd::do_reset() throw() {

    if (a_threadRunning) {
      std::string s_warn = anna::functions::asString("Thread still in progress: deprecating step %d for Test Case %llu", getNumber(), a_testCase->getId());
      LOGWARNING(anna::Logger::warning(s_warn, ANNA_FILE_LOCATION));
      a_threadDeprecated = true;
    }
//  if (a_threadRunning) {
//    std::string s_warn = anna::functions::asString("Thread still in progress: killing child pid %d within step %d for Test Case %llu", a_childPid, getNumber(), a_testCase->getId());
//    LOGWARNING(anna::Logger::warning(s_warn, ANNA_FILE_LOCATION));
//    kill (a_childPid, SIGKILL);
//  }

  a_resultCode = -2;
  a_errorMsg = "";
  //a_output = "";
  a_childPid = -1;
}

anna::xml::Node* TestStepIpLimit::asXML(anna::xml::Node* parent)
throw() {
  anna::xml::Node* result = TestStep::asXML(parent);
  std::string limit = (a_ipLimit != UINT_MAX) ? anna::functions::asString(a_ipLimit) : "<no limit>";
  result->createAttribute("IpLimit", limit);

  return result;
}

bool TestStepIpLimit::do_execute() throw() {
  TestManager::instantiate().setInProgressLimit(a_ipLimit);
  complete();
  return true; // go next
}

void TestStepIpLimit::do_complete() throw() {
  if (TestManager::instantiate().getDumpStdout()) {
    std::string limit = (a_ipLimit != UINT_MAX) ? anna::functions::asString(a_ipLimit) : "<no limit>";
    std::cout << "Executed IpLimit Test Step (value = " << limit << ")" << std::endl;
  }
}

