// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <ctype.h>

#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <libxml/parser.h>

#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/tracing/TraceFunction.hpp>

#include <anna/xml/XPath.hpp>
#include <anna/xml/internal/sccs.hpp>
#include <anna/xml/Document.hpp>

using namespace std;
using namespace anna;

xml::XPath::XPath(const char* name) {
  a_root = new Node(name);
  sccs::activate();
}

/*
 * Segn la implementacin de
 * http://www.xmlsoft.org/examples/xpath1.c
 *
 * Los nodos se recuperan mediante los mtodos node_begin, node_end y node
 */
void xml::XPath::apply(const Document& document, const char* expression, const int mode)
throw(RuntimeException) {
  a_mode = mode;
  ConfigSkeleton configSkeleton(&XPath::callbackApply);

  if((mode & Mode::Namespace) != 0)
    configSkeleton.initialize = &XPath::initializeNamespaces;

  skeleton(document, expression,  configSkeleton);
}

bool xml::XPath::match(const Document& document, const char* expression, const int mode)
throw(RuntimeException) {
  a_result = false;
  ConfigSkeleton configSkeleton(&XPath::callbackMatch);

  if((mode & Mode::Namespace) != 0)
    configSkeleton.initialize = &XPath::initializeNamespaces;

  skeleton(document, expression, configSkeleton);
  LOGDEBUG(
    string msg("xml::XPath::match | Expression: ");
    msg += expression;
    msg += anna::functions::asText(" | Result: ", a_result);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  return a_result;
}

const xml::Node* xml::XPath::find(const Document& document, const char* expression, const int mode, const Exception::Mode::_v emode)
throw(RuntimeException) {
  apply(document, expression, mode);
  Node::const_child_iterator ii = node_begin();

  if(ii == node_end()) {
    if(emode == Exception::Mode::Throw) {
      string msg("xml::XPath::find | Expression: ");
      msg += expression;
      msg += " | Does not match any xml::Node";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    } else if(emode == Exception::Mode::Trace) {
      string msg("xml::XPath::find | Expression: ");
      msg += expression;
      msg += " | Does not match any xml::Node";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    }

    return NULL;
  }

  if(size() > 1) {
    string msg("xml::XPath::find | Expression: ");
    msg += expression;
    msg += functions::asString(" | Ignores %d xml::Node's", size());
    Logger::warning(msg, ANNA_FILE_LOCATION);
  }

  return XPath::node(ii);
}

void xml::XPath::skeleton(const Document& document, const char* expression, xml::XPath::ConfigSkeleton& config)
throw(RuntimeException) {
  LOGDEBUG(
    string msg("xml::XPath::skeleton | Document: ");
    msg += functions::asString(document.getContent());
    msg += " | Expression: ";
    msg += expression;
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  if(document.a_handle == NULL)
    throw RuntimeException("xml::XPath::pattern | xml::Document had not been initialized", ANNA_FILE_LOCATION);

  setupEncoding(document.a_handle);
  a_root->clear();
  xmlXPathContextPtr context = NULL;
  xmlXPathObjectPtr object = NULL;

  try {
    if((context = xmlXPathNewContext(document.a_handle)) == NULL)
      throw RuntimeException("Can not create XPath context", ANNA_FILE_LOCATION);

    if(config.initialize != NULL)
      (this->*config.initialize)(context, document);

    if((object = xmlXPathEvalExpression(BAD_CAST expression, context)) == NULL) {
      string msg("xml::XPath::skeleton | Expression: ");
      msg += expression;
      msg += " | Unable to evaluate expression";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    (this->*config.callback)(object->nodesetval);
    xmlXPathFreeObject(object);
    xmlXPathFreeContext(context);
  } catch(RuntimeException& ex) {
    if(object != NULL)
      xmlXPathFreeObject(object);

    if(context != NULL)
      xmlXPathFreeContext(context);

    throw;
  }
}

/*
 * Ha que tener en cuenta que puede haber definiciones de NS en cualquier nodo
 * lo que nos obliga a recorrerlo en profundidad.
 */
void xml::XPath::initializeNamespaces(_xmlXPathContext* context, const Document& document)
throw(RuntimeException) {
  xmlNode* root;

  if((root = xmlDocGetRootElement(document.a_handle)) == NULL)
    throw RuntimeException("Error interpreting the XML document", ANNA_FILE_LOCATION);

  // Incorpora la definicin de los namespaces usados en el documento y los define en el contexto XPath
  forwardNamespaces(context, root);
}

void xml::XPath::forwardNamespaces(_xmlXPathContext* context, _xmlNode* xmlNode)
throw(RuntimeException) {
  while(xmlNode != NULL) {
    if(xmlNode->type == XML_ELEMENT_NODE) {
      for(xmlNs* ns = xmlNode->nsDef; ns != NULL; ns = ns->next) {
        a_root->createNamespace((const char*) ns->prefix, (const char*) ns->href);
        xmlXPathRegisterNs(context, ns->prefix, ns->href);
        LOGDEBUG(
          string msg("xml::XPath::forwardNamespaces | Prefix: ");
          msg += (const char*) ns->prefix;
          msg += " | HRef: ";
          msg += (const char*) ns->href;
          Logger::debug(msg, ANNA_FILE_LOCATION);
        );
      }

      forwardNamespaces(context, xmlNode->children);
    }

    xmlNode = xmlNode->next;
  }
}

void xml::XPath::callbackApply(const _xmlNodeSet* nodes) {
  /* Si no hay nodos que cumplan con la expresin */
  if(nodes == NULL)
    return;

  Node* node;
  xmlNodePtr xmlNode;

  for(int ii = 0, maxii = nodes->nodeNr; ii < maxii; ii ++) {
    xmlNode = nodes->nodeTab [ii];

    if(xmlNode->type != XML_ELEMENT_NODE)
      continue;

    /* Slo analiza el primer nivel de dependencia de cada uno de los nodos seleccionados */
    node = a_root->createChild((const char*) xmlNode->name);
    attributes(node, xmlNode->properties);

    if(xmlNode->ns != NULL)
      node->setNamespace(a_root->namespace_find((const char*) xmlNode->ns->prefix));

    if((a_mode & Mode::Full) != 0)
      children(node, xmlNode->children);
    else
      text(node, xmlNode->children);

    LOGDEBUG(Logger::debug(node->asString(), ANNA_FILE_LOCATION));
  }
}

void xml::XPath::callbackMatch(const _xmlNodeSet* nodes) {
  a_result = (nodes == NULL) ? false : (nodes->nodeNr > 0);
}

void xml::XPath::text(Node* node, xmlNode* xmlNode)
throw(RuntimeException) {
  bool isSeparator;
  const char* w;

  while(xmlNode != NULL) {
    switch(xmlNode->type) {
    case XML_TEXT_NODE:
      w = (const char*) xmlNode->content;
      isSeparator = true;

      while(*w != 0) {
        if(isspace(*w) == false) {
          isSeparator = false;
          break;
        }

        w ++;
      }

      if(isSeparator == false)
        node->createText(decode(xmlNode->content));

      return;

      default:
       ;
    }

    xmlNode = xmlNode->next;
  }
}



